#!/usr/bin/env python
#
# $Id: MapStack.py,v 1.1 2003/01/26 19:12:11 doughellmann Exp $
#
# Copyright 2003 Doug Hellmann.
#
#
#                         All Rights Reserved
#
# Permission to use, copy, modify, and distribute this software and
# its documentation for any purpose and without fee is hereby
# granted, provided that the above copyright notice appear in all
# copies and that both that copyright notice and this permission
# notice appear in supporting documentation, and that the name of Doug
# Hellmann not be used in advertising or publicity pertaining to
# distribution of the software without specific, written prior
# permission.
#
# DOUG HELLMANN DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
# INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN
# NO EVENT SHALL DOUG HELLMANN BE LIABLE FOR ANY SPECIAL, INDIRECT OR
# CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
# OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
# NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#

"""Implements a dictionary-like class to provide namespace and context stacking.

Based on the MultiMapping extension class found in Zope.
"""

__rcs_info__ = {
    #
    #  Creation Information
    #
    'module_name'  : '$RCSfile: MapStack.py,v $',
    'rcs_id'       : '$Id: MapStack.py,v 1.1 2003/01/26 19:12:11 doughellmann Exp $',
    'creator'      : 'Doug Hellmann <doug@hellfly.net>',
    'project'      : 'HappyDoc',
    'created'      : 'Sun, 19-Jan-2003 07:45:43 EST',

    #
    #  Current Information
    #
    'author'       : '$Author: doughellmann $',
    'version'      : '$Revision: 1.1 $',
    'date'         : '$Date: 2003/01/26 19:12:11 $',
}
try:
    __version__ = __rcs_info__['version'].split(' ')[1]
except:
    __version__ = '0.0'

#
# Import system modules
#
import unittest

#
# Import Local modules
#


#
# Module
#

class MapStack:
    """Implements a dictionary-like class to provide namespace and context stacking.

    Based on the MultiMapping extension class found in Zope.
    """

    def __init__(self, *dicts):
        """Construct a new MapStack.

        Pass some existing mappings to be pushed onto the stack.
        """
        #print 'MapStack', dicts
        self.stack = []
        for d in dicts:
            self.push(d)
        return

    def push(self, newDict):
        """Add a new context to the stack.
        """
        self.stack.insert(0, newDict)
        return
    _push = push

    def pop(self):
        """Remove and return the top context.
        """
        if self.stack:
            top = self.stack[0]
        else:
            top = None
        self.stack = self.stack[1:]
        return top
    _pop = pop

    def __getitem__(self, key):
        for map in self.stack:
            if map.has_key(key):
                return map[key]
        raise KeyError(key)

    def dump(self):
        for map in self.stack:
            print map.keys()
        return

    def __setitem__(self, name, value):
        #print 'MapStack.setitem(%s, %s)' % (name, value)
        if not self.stack:
            raise ValueError('Stack empty')
        top = self.stack[0]
        top[name] = value
        return

    def has_get(self, key):
        try:
            val = self[key]
        except KeyError:
            has = 0
            val = None
        else:
            has = 1
        #print 'has_get(%s) -> (%s, %s)' % (key, has, val)
        #if not has:
        #    self.dump()
        return (has, val)


class MapStackTest(unittest.TestCase):

    def testEmpty(self):
        ms = MapStack()
        try:
            anything = ms['anything']
        except KeyError:
            pass
        else:
            self.fail('Should have raised KeyError.  Got: %s' % anything)
        return

    def testOneLevel(self):
        ms = MapStack()
        ms.push({'one':1, 'two':2})
        try:
            one = ms['one']
        except KeyError:
            self.fail('Should have found one')
            
        try:
            two = ms['two']
        except KeyError:
            self.fail('Should have found two')
        return
    
    def testTwoLevels(self):
        ms = MapStack()
        ms.push({'one':1})
        ms.push({'two':2})
        try:
            one = ms['one']
        except KeyError:
            self.fail('Should have found one')
            
        try:
            two = ms['two']
        except KeyError:
            self.fail('Should have found two')
        return
    
    def testTwoLevelsWithOverride(self):
        ms = MapStack()
        ms.push({'one':1})
        ms.push({'one':3, 'two':2})
        try:
            one = ms['one']
        except KeyError:
            self.fail('Should have found one')
        else:
            self.failUnlessEqual(one, 3)
            
        try:
            two = ms['two']
        except KeyError:
            self.fail('Should have found two')
        return
    
    def testTwoLevelsWithPop(self):
        ms = MapStack()
        ms.push({'one':1})
        try:
            one = ms['one']
        except KeyError:
            self.fail('Should have found one')
        else:
            self.failUnlessEqual(one, 1)
            
        ms.push({'one':3, 'two':2})
        try:
            one = ms['one']
        except KeyError:
            self.fail('Should have found one')
        else:
            self.failUnlessEqual(one, 3)
            
        try:
            two = ms['two']
        except KeyError:
            self.fail('Should have found two')

        removed = ms.pop()
        try:
            one = ms['one']
        except KeyError:
            self.fail('Should have found one')
        else:
            self.failUnlessEqual(one, 1)

        try:
            two = ms['two']
        except KeyError:
            pass
        else:
            self.fail('Should have gotten a KeyError')
        return
    
        
    
if __name__ == '__main__':
    unittest.main()
